import os

CACHE_PATH = os.path.expanduser('~/.teilice')
TESSCUT_PATH = os.path.join(CACHE_PATH, 'tesscut')
NEARBY_PATH  = os.path.join(CACHE_PATH, 'nearby')
DOWNSCRIPT_PATH = os.path.join(CACHE_PATH, 'download_scripts')
