#!/usr/bin/env python3
from distutils.core import setup

setup(
    name         = 'teilice',
    version      = '0.1',
    description  = '',
    author       = 'Liang Wang',
    author_email = 'wang.leon@gmail.com',
    license      = 'Apache-2.0',
    packages     = [
                    'teilice',
                    ],

    )
